#
#  Name:    optimize_auto_func.tcl
#
#  Purpose: This file contains utility functions for the 'optimize_auto'
#           program.
#
#  Author:  William Lavender
#
#-------------------------------------------------------------------------
#
#  Copyright 2001-2002 Illinois Institute of Technology
#
#  See the file "LICENSE" for information on usage and redistribution
#  of this file, and a DISCLAIMER OF ALL WARRANTIES.
#

namespace eval MxOptimizeAuto {
	variable current_optimization_name ""
	variable default_button_background
	variable motor_name_list
	variable mxdir
	variable logfile_name
	variable test
	variable use_gui 0
	variable user_interrupt 0
	variable user_script

	set motor_name_list [ list ]
}

proc move_motor { record_list motor_name destination } {

	if [ catch { $record_list get_record $motor_name } motor ] {

		log_puts \
		  "Cannot find the motor '$motor_name' in the MX database."

		return -code error \
		  "Cannot find the motor '$motor_name' in the MX database."
	}

	set units [ $motor get_field "units" ]

	if [ catch { $motor move_absolute $destination } ] {

		log_puts \
	"The attempt to move '$motor_name' to $destination $units failed."

		return -code error \
	"The attempt to move '$motor_name' to $destination $units failed."
	}

	while {1} {
		if [ catch { $motor is_busy } busy ] {
			log_puts \
			"Cannot determine motion status of motor '$motor_name'."

			return -code error \
			"Cannot determine motion status of motor '$motor_name'."
		}
		if { $busy == 0 } {
			return
		}
		update idletasks

		after 100
	}
}

proc optimize_to_peak { record_list motor_name } {
	optimize_main $record_list $motor_name "fit_to_peak"
}

proc optimize_deriv_to_peak { record_list motor_name } {
	optimize_main $record_list $motor_name "fit_deriv_to_peak"
}

proc optimize_to_edge { record_list motor_name } {
	optimize_main $record_list $motor_name "fit_to_edge"
}

proc optimize_main { record_list motor_name fit_type } {

	#
	# Find the motor record.
	#

	set MxOptimizeAuto::current_optimization_name $motor_name

#	set text    "MxOptimizeAuto::current_optimization_name = "
#	append text "'$MxOptimizeAuto::current_optimization_name'"
#	log_puts $text

	if [ catch { $record_list get_record $motor_name } motor ] {

		log_puts \
		"Cannot find the motor '$motor_name' in the MX database."

		return -code error \
		"Cannot find the motor '$motor_name' in the MX database."
	}

	#
	# Save the old motor position.
	#

	set old_position [ $motor get_position ]

	set units [ $motor get_field "units" ]

	update_motor_position $motor_name old $old_position

	#
	# Find the optimization scan.
	#

	set scan_name "${motor_name}_scan"

	if [ catch { $record_list get_record $scan_name } scan ] {

		log_puts \
	"Cannot find the optimization scan '$scan_name' in the MX database."

		return -code error \
	"Cannot find the optimization scan '$scan_name' in the MX database."
	}

	#
	# Calculate the scan range.
	#

	set step_size [ $scan get_field "step_size" ]
	set num [ $scan get_field "num_measurements" ]

	set scan_range [ expr ( ($num - 1) * $step_size ) ]

	#
	# Change the scan start position so that the scan range is centered
	# on the original position of the motor.
	#

	set scan_start [ expr ( $old_position - 0.5 * $scan_range ) ]

	$scan set_field "start_position" $scan_start

	#
	# Run the optimization scan.
	#

	modify_status $motor_name yellow "Optimization scan in progress"

	setup_scan_abort_button

	if [ catch { $scan perform_scan } ] {
		log_puts "The optimization scan '$scan_name' failed."
		log_puts "Moving '$motor_name' back to $old_position $units"

		move_motor $record_list $motor_name $old_position

		return -code error "Optimization scan '$scan_name' was aborted."
	}

	set scan_datafile_name [ $scan get_field "datafile_name" ]

	if { $fit_type == "fit_to_peak" } {
		set fit_datafile_name $scan_datafile_name

	} elseif { $fit_type == "fit_to_edge" } {
		set fit_datafile_name $scan_datafile_name

	} elseif { $fit_type == "fit_deriv_to_peak" } {
                # Construct the name of a temporary file to write
                # the derivative values to.
 
                switch $tcl_platform(platform) {
                    unix {
                        set tmpdir "/tmp"
                    } default {
                        set tmpdir [pwd]
                        catch { set tmpdir $env(TMP) }
                        catch { set tmpdir $env(TEMP) }
                    }
                }
 
                set temp_datafile_name \
			[ file join $tmpdir "optimize_auto.[ pid ]" ]
 
                set fit_datafile_name $temp_datafile_name
 
                # Differentiate the data.

		modify_status $motor_name yellow "Differentiation in progress"
 
                set command \
                  "opt_differentiate $scan_datafile_name $temp_datafile_name"
 
                log_puts "Running differentiation command '$command'"
 
                set status [ catch { eval exec $command } ]
 
                if { $status != 0 } {
			modify_status $motor_name red "Differentiation failed."

			log_puts "The differentiation program failed."
			log_puts \
		"Moving back to the original position of $old_position $units."

			move_motor $record_list $motor_name $old_position

			catch { file delete $temp_datafile_name }

			return -code error \
			"The differentiation program failed."
		}
	} else {
		modify_status $motor_name red "Unknown fit type"

		log_puts "Unknown curve fitting type $fit_type"

		log_puts \
		"Moving back to the original position of $old_position $units."

		move_motor $record_list $motor_name $old_position

		return -code error "Unknown curve fitting type $fit_type"
	}

	#
	# Perform the fit with an external script.
	#

	modify_status $motor_name yellow "Fitting in progress"

	if { $fit_type == "fit_to_edge" } {
		set program "optimize_step_fitting"
	} else {
		set program "optimize_peak_fitting"
	}

	set num_lines [ eval exec "wc -l $fit_datafile_name" ]

	scan $num_lines "%d" num_lines

	#

	set command "$program $num_lines $fit_datafile_name"

	log_puts "Running fit command '$command'."

	#
	# Run the fit command.
	#

	set status [ catch { eval exec $command } result ]

	if [ info exists temp_datafile_name ] {
		catch { file delete $temp_datafile_name }
	}

	if { $status != 0 } {
		modify_status $motor_name red "Fit crashed"

		log_puts "The fitting program crashed with status = '$status'"
		log_puts \
		"Moving back to the original position of $old_position $units."

		move_motor $record_list $motor_name $old_position

		return -code error \
		"The fitting program crashed with status = '$status'"
	}
	if { $result == "" } {
		modify_status $motor_name red "No fit returned"

		log_puts "The fitting program did not return a result."
		log_puts \
		"Moving back to the original position of $old_position $units."

		move_motor $record_list $motor_name $old_position

		return -code error \
		"The fitting program did not return a result."
	}

	#
	# Convert the result to lower case and replace any D format characters
	# with e.
	#

	set result [ string tolower $result ]

	set length [ string length $result ]

	for { set i 0 } { $i < $length } { incr i } {

		set thischar [ string index $result $i ]

		if { $thischar == "d" } {

			set thischar "e"
		}

		append new_result $thischar
	}

#	log_puts "new_result = '$new_result'"

	#
	# Did the fit work?
	#

	scan $new_result "%d %g %d %g %g %g %g %g" \
		fit_status euclidean_norm algorithm_info simple_position \
		a1 a2 a3 a4

#	log_puts "fit_status = '$fit_status'"
#	log_puts "euclidean_norm = '$euclidean_norm'"
#	log_puts "algorithm_info = '$algorithm_info'"
#	log_puts "max_position = '$max_position'"
#	log_puts "a1 = '$a1'"
#	log_puts "a2 = '$a2'"
#	log_puts "a3 = '$a3'"
#	log_puts "a4 = '$a4'"

	if { $fit_type == "fit_to_edge" } {
		set new_position $simple_position
	} else {
		set new_position $a3
	}

	if { $fit_status == 0 } {
		modify_status $motor_name red "Fit failed"

		log_puts \
"Optimization fit failed.  Moving to old position $old_position $units"

		move_motor $record_list $motor_name $old_position
		exit 1
	}

	#
	# If the fit worked, move the motor to the new position.
	#

	modify_status $motor_name springgreen "Moving to optimized position"

	log_puts \
"Moving motor '$motor_name' to the fitted position of $new_position $units"

	move_motor $record_list $motor_name $new_position

	modify_status $motor_name limegreen "Optimization succeeded"

	update_motor_position $motor_name new $new_position

	set MxOptimizeAuto::current_optimization_name ""

#	set text    "MxOptimizeAuto::current_optimization_name = "
#	append text "'$MxOptimizeAuto::current_optimization_name'"
#	log_puts $text

	return
}

proc construct_file_name { oldname } {

	if { [ file tail $oldname ] == $oldname } {

		if { $MxOptimizeAuto::test } {
			set file_name $oldname
		} else {
			set file_name [ file join \
			$MxOptimizeAuto::mxdir "etc" "optimize_auto" $oldname ]
		}
	} else {
		set file_name $oldname
	}

#	puts "file_name = '$file_name'"

	return $file_name
}

proc optimize_auto_output_function { message } {
	log_puts $message
	update
}

proc optimize_auto_info_dialog_function { text_prompt \
					gui_prompt button_label } {

	bell
	tk_dialog .optimize_auto_info_dialog "MX Info" \
		$gui_prompt info 0 $button_label
}

proc optimize_auto_user_interrupt_function {} {

	update

	set interrupt_occurred $MxOptimizeAuto::user_interrupt

	set MxOptimizeAuto::user_interrupt 0

	return $interrupt_occurred
}

proc log_puts { message } {
	if { $MxOptimizeAuto::use_gui } {
		set textwindow [ .message_log childsite ]

		$textwindow configure -state normal

		.message_log insert end "$message\n"

		$textwindow configure -state disabled

		.message_log yview moveto 1.0

		update idletasks
	} else {
		puts $message
	}
}

proc logfile_puts { message } {

	# Log file messages are only recorded if $MxOptimizeAuto::logfile_name
	# has a length greater than zero.

	if { $MxOptimizeAuto::logfile_name == "" } {
		return
	}

	#
	# Construct a timestamp to write to the log file.
	#

	set current_time [ clock seconds ]

	set timestamp [ clock format $current_time -format \
		"%d/%m/%Y %H:%M:%S " ]

	#
	# Finish by writing the message to the log file.
	#

	set logfile [ open $MxOptimizeAuto::logfile_name "a" ]

	puts $logfile "$timestamp $message"

	close $logfile

	return
}

proc create_message_log_window {} {

	set window .message_log

        #
        #  Create the window.
        #

        iwidgets::scrolledtext $window -labeltext "Message Log"

        $window configure -visibleitems 40x10
        $window configure -hscrollmode dynamic
        $window configure -vscrollmode dynamic

        pack $window -expand yes -fill both

        #
        #  Install handlers to redirect the output of mx_info(), mx_error()
        #  and MX_DEBUG() to this window.
        #

	Mx::set_all_output_functions  optimize_auto_output_function

	#
	#  Install user interrupt and dialog handlers.  This is necessary
	#  since otherwise the program will block on tty input if run in
	#  the background.
	#

        Mx::set_user_interrupt_function  optimize_auto_user_interrupt_function

	Mx::set_info_dialog_function  optimize_auto_info_dialog_function

	log_puts "*** Auto Optimization GUI Initializing ***"
}

proc add_main_window_to_gui { title } {

	frame .title
	pack .title -before .message_log

	label .title.text -text $title
	pack .title.text

	frame .sep1 -height 2 -borderwidth 1 -relief sunken
	pack .sep1 -expand yes -fill x -pady 5 -before .message_log

	frame .main
	pack .main -padx 10 -before .message_log

	label .main.blank_header1
	label .main.blank_header2
	label .main.old_header -text "Old Position"
	label .main.new_header -text "New Position"

	grid .main.blank_header1 .main.blank_header2 \
		.main.old_header .main.new_header -row 0 -sticky w

}

proc add_motor_to_gui { row record_list motor_name label } {

	set statuswidth 30

	set motor [ $record_list get_record $motor_name ]

	set thisfield .main.${motor_name}

	label ${thisfield}_name -text "$label"
	label ${thisfield}_status -width $statuswidth -text "Not Yet Optimized"
	label ${thisfield}_old_position -text [ $motor get_position ]
	label ${thisfield}_new_position -text "---"

	grid ${thisfield}_name         -row $row -column 0 -sticky w
	grid ${thisfield}_status       -row $row -column 1 -sticky w
	grid ${thisfield}_old_position -row $row -column 2 -sticky e
	grid ${thisfield}_new_position -row $row -column 3 -sticky e

	lappend MxOptimizeAuto::motor_name_list $motor_name
}

proc add_buttons_to_gui {} {

	frame .sep2 -height 2 -borderwidth 1 -relief sunken
	pack .sep2 -expand yes -fill x -pady 8 -before .message_log

	frame .bottom
	pack .bottom -before .message_log

	button .bottom.button -text "---"
	pack .bottom.button -side left

	set MxOptimizeAuto::default_button_background \
		[ .bottom.button cget -background ]

	label .bottom.sep -width 5
	pack .bottom.sep -side left -after .bottom.button

	button .bottom.dismiss -text Dismiss -command exit
	pack .bottom.dismiss -side left -after .bottom.sep

	frame .sep3 -height 2 -borderwidth 1 -relief sunken
	pack .sep3 -expand yes -fill x -pady 8 -before .message_log

}

proc update_motor_position { motor_name type position } {
	if { $MxOptimizeAuto::use_gui == 0 } {
		return
	}

	set window .main.${motor_name}_${type}_position

	$window configure -text $position

	update idletasks
}

proc modify_status { motor_name color status } {
	if { $MxOptimizeAuto::use_gui == 0 } {
		return
	}

	set window .main.${motor_name}_status

	$window configure -foreground $color
	$window configure -text $status

	update
}

proc setup_start_button { record_list } {
	if { $MxOptimizeAuto::use_gui == 0 } {
		return
	}

	.bottom.button configure -background \
			$MxOptimizeAuto::default_button_background
	.bottom.button configure -activebackground \
			$MxOptimizeAuto::default_button_background

	.bottom.button configure -text "Start" -background limegreen
	.bottom.button configure -command [ itcl::code \
					scan_start_handler $record_list ]
}

proc setup_scan_abort_button {} {
	if { $MxOptimizeAuto::use_gui == 0 } {
		return
	}

	.bottom.button configure -background red
	.bottom.button configure -activebackground red
	.bottom.button configure -text "Abort Scan"
	.bottom.button configure -command scan_abort_handler
}

proc hide_dismiss {} {
	pack forget .bottom.sep
	pack forget .bottom.dismiss
}

proc show_dismiss {} {
	pack .bottom.sep -side left -after .bottom.button
	pack .bottom.dismiss -side left -after .bottom.sep
}

proc scan_start_handler { record_list } {

	set name_list $MxOptimizeAuto::motor_name_list

	foreach name $name_list {
		set motor [ $record_list get_record $name ]

		update_motor_position $name old [ $motor get_position ]
		update_motor_position $name new "---"

		modify_status $name black "Not Yet Optimized"
	}

	hide_dismiss

	if [ catch { main $record_list } ] {

		log_puts \
	"*** User script '$MxOptimizeAuto::user_script' ended abnormally. ***"

		if [ info exists errorInfo ] {
			log_puts "Traceback = '$errorInfo'"
		}

		set name $MxOptimizeAuto::current_optimization_name

		if { $name != "" } {
			modify_status $name red "Optimization scan failed"
		}
	}

	setup_start_button $record_list

	show_dismiss
}

proc scan_abort_handler {} {
	log_puts "Scan aborted."

	set MxOptimizeAuto::user_interrupt 1
}

